import os
import tempfile


class Editor:
    def __init__(self, content: str) -> None:
        self._content = content

    def edit(self) -> None:
        new_file, filename = tempfile.mkstemp(".yaml")
        try:
            with os.fdopen(new_file, 'w') as tmp:
                tmp.write(self._content)

            os.system(f"$EDITOR {filename} || nvim {filename} || vim {filename}")

            with open(filename, 'r') as tmp:
                return tmp.read()
        finally:
            os.remove(filename)




from __future__ import annotations

from enum import Enum
from typing import List, Optional
from datetime import datetime

from pydantic import BaseModel, Field


class GlobalConfig(BaseModel):
    all_tasks_files: List[str] = []
    items_counter: int = 0

    def increase_items_counter(self) -> None:
        self.items_counter += 1


class ItemState(str, Enum):
    NEW = "NEW"
    TODO = "TODO"
    DONE = "DONE"


class Item(BaseModel):
    id: int
    created_at: datetime = Field(default_factory=datetime.now)
    updated_at: Optional[datetime] = None
    state: ItemState = ItemState.NEW
    tags: List[str] = []
    title: str

    class Config:
        use_enum_values = True


class Data(BaseModel):
    created_at: datetime = Field(default_factory=datetime.now)
    items: List[Item] = []



import json
import os
from pathlib import Path

from .models import Data, GlobalConfig


class DataLoader:
    def __init__(self, path: str = None) -> None:
        self._filename = path or Path(os.getcwd()) / ".tasks.json"
        self._global_config_filename = Path(os.path.expanduser("~")) / ".tasks.config.json"


    def exists(self) -> bool:
        return os.path.exists(self._filename)

    def init(self) -> None:
        with open(self._filename, "w+") as f:
            f.write(Data().json())

    def load(self) -> Data:
        with open(self._filename, "r") as f:
            return Data(**json.load(f))

    def create_global_config_if_not_exists(self) -> None:
        if not os.path.exists(self._global_config_filename):
            with open(self._global_config_filename, "w+") as f:
                f.write(GlobalConfig().json())

    def load_global_config(self) -> GlobalConfig:
        with open(self._global_config_filename, "r") as f:
            return GlobalConfig(**json.load(f))

    def save_global_config(self, global_config: GlobalConfig) -> None:
        with open(self._global_config_filename, "w+") as f:
            f.write(global_config.json())

    def drop(self) -> None:
        os.remove(self._filename)

    def save(self, data: Data) -> None:
        with open(self._filename, "w+") as f:
            f.write(data.json())



from typing import List

from rich import print
from rich.table import Table
from rich.console import Console

from .models import Data, Item
from .time import get_time_difference


class ListView:
    def __init__(self, items: List[Item]) -> None:
        self._items = items

    def print(self) -> str:
        console = Console()
        table = Table()

        table.add_column("ID", justify="right", style="cyan", no_wrap=True)
        table.add_column("Title", style="magenta")
        table.add_column("Tags", style="red")
        table.add_column("Status")
        table.add_column("Created", justify="right", style="green")

        for item in self._items:
            table.add_row(*self.get_row(item))

        console.print(table)

    def get_row(self, item: Item) -> str:
        return (f"#{item.id}", item.title, self.get_tags(item.tags), item.state, get_time_difference(item.created_at))

    def get_tags(self, tags: List[str]) -> str:
        return " ".join(map(lambda i: f"+{i}", tags))



import click
from rich import print

from .data import DataLoader
from .view import ListView
from .logic import Logic


class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail("Too many matches: %s" % ", ".join(sorted(matches)))


@click.group(cls=AliasedGroup)
def cli() -> None:
    pass


@click.command()
def init() -> None:
    data_loader = DataLoader()
    data_loader.create_global_config_if_not_exists()

    if data_loader.exists():
        print("Init file already exists :thumbs_up:")
    else:
        print("Tasks succesfully initialized :thumbs_up:")
        data_loader.init()


@click.command()
def list_all() -> None:
    data_loader = DataLoader()
    data = data_loader.load()
    logic = Logic(data)
    items = logic.list_all()
    view = ListView(items)
    view.print()


@click.command()
@click.confirmation_option(prompt="Are you sure you want to drop all your tasks there?")
def drop() -> None:
    data_loader = DataLoader()
    data_loader.drop()
    print("Local tasks lost forever :sad_panda:")


@click.command()
@click.option("-s", "--state", default="NEW", type=str)
@click.argument("title", type=str, nargs=-1)
def add(state: str, title: str) -> None:
    data_loader = DataLoader()
    data = data_loader.load()
    global_config = data_loader.load_global_config()

    logic = Logic(data)
    new_data = logic.add(" ".join(title), global_config.items_counter, state)
    data_loader.save(new_data)

    global_config.increase_items_counter()
    data_loader.save_global_config(global_config)

    print("Task saved", ":vampire:")


@click.command()
@click.argument("task_id", type=int)
def remove(task_id: str) -> None:
    data_loader = DataLoader()
    data = data_loader.load()

    logic = Logic(data)
    new_data = logic.remove(task_id)
    data_loader.save(new_data)

    print("Task removed :sad_panda:")


@click.command()
@click.argument("task_id", type=int)
def edit(task_id: str) -> None:
    data_loader = DataLoader()
    data = data_loader.load()

    logic = Logic(data)
    new_data = logic.edit(task_id)
    data_loader.save(new_data)

    print("Task removed :sad_panda:")


cli.add_command(add)
cli.add_command(edit)
cli.add_command(remove)
cli.add_command(list_all)
cli.add_command(init)
cli.add_command(drop)



import json
import yaml
from typing import List

from .editor import Editor
from .models import Data, Item, ItemState


class Logic:
    def __init__(self, data: Data) -> None:
        self._data = data

    def list_all(self) -> List[Item]:
        return self._data.items

    def add(self, title: str, new_id: int, state: str) -> Data:
        self._data.items.append(Item(id=new_id, title=title, state=ItemState(state)))
        return self._data

    def remove(self, task_id: int) -> None:
        self._data.items = [i for i in self._data.items if i.id != task_id]
        return self._data

    def edit(self, task_id: int) -> Item:
        item_id = next(i for i, item in enumerate(self._data.items) if item.id == task_id)
        item = self._data.items[item_id]
        editor = Editor(yaml.dump(item.dict(include={"title", "state", "tags"})))
        new_item_dict = yaml.safe_load(editor.edit())
        self._data.items[item_id] = Item(**{**item.dict(), **new_item_dict})
        return self._data
